use crate::model::Modelable;

/// A database entity, aka a struct representing a row in a table
pub trait Entity: 'static + for<'de> serde::Deserialize<'de> + serde::Serialize {
    type ID: EntityID;
    fn table_name() -> &'static str;
    fn column_count() -> usize
    where
        Self: Sized;
    fn columns() -> &'static [&'static dyn EntityColumn<Entity = Self>];
    fn values(&self) -> Vec<&dyn Modelable>;

    fn build_from(stmt: &sqlite::Statement) -> sqlite::Result<Self>
    where
        Self: Sized;

    // fn foreign_keys() -> &'static [&'static dyn EntityForeignKey<Local = dyn EntityColumn<Entity = Self>, Foreign = dyn EntityColumn<Entity = ...>>];
}

/// Trait representing the columns of a database entity
pub trait EntityColumn: 'static + Send + Sync {
    type Entity: Entity;

    fn sql_type(&self) -> &'static str;
    fn index(&self) -> usize;
    fn name(&self) -> &'static str;

    fn fk_table_name(&self) -> Option<&'static str>;
    fn fk_column_name(&self) -> Option<&'static str>;
}

/// Trait for entity IDs in the database
pub trait EntityID: 'static + std::fmt::Debug + Copy + Modelable {
    type Entity: Entity;
    fn from_raw_id(raw: i64) -> Self;
    fn raw_id(&self) -> i64;
}

/// Trait for a foreign key relationship
pub trait EntityForeignKey {
    type Local: EntityColumn;
    type Foreign: EntityColumn;
}

/// Trait for an index over a column
pub trait Index {
    type IndexedEntity: Entity;

    fn index_name() -> &'static str
    where
        Self: Sized;
    fn columns() -> &'static [&'static dyn EntityColumn<Entity = Self::IndexedEntity>]
    where
        Self: Sized;
    fn unique() -> bool
    where
        Self: Sized;
}
